from abc import ABC, abstractmethod
from typing import List, Tuple

import torch
from torch import Tensor
from torch.nn import Module
from torch.optim.lr_scheduler import _LRScheduler
from torch.optim.optimizer import Optimizer


class BaseModule(ABC, Module):
    def __init__(
            self,
        ):
        super().__init__()

    @abstractmethod
    def forward(
            self,
            x: Tensor,
        ) -> Tensor:
        raise NotImplementedError

    @abstractmethod
    def training_step(
            self,
            batch: Tensor,
            batch_idx: int,
            optimizer_idx: int,
        ) -> Tensor:
        raise NotImplementedError

    def training_step_end(
            self,
            batch_idx: int,
        ):
        pass

    def training_epoch_end(
            self,
            epoch_idx: int,
        ):
        pass

    @abstractmethod
    def validation_step(
            self,
            batch: Tensor,
            batch_idx: int,
        ) -> Tensor:
        raise NotImplementedError

    def validation_step_end(
            self,
            batch_idx: int,
        ):
        pass

    def validation_epoch_end(
            self,
            epoch_idx: int,
        ):
        pass

    @abstractmethod
    def configure_optimizers(
            self,
        ) -> Tuple[List[Optimizer], List[_LRScheduler]]:
        raise NotImplementedError




from typing import Tuple

from torch.utils.data import DataLoader
from torchvision.datasets import CelebA as CelebADataset
from torchvision.transforms import (
    Compose,
    Normalize,
    Resize,
    ToTensor,
)

from mr.datamodules.base_datamodule import BaseDataModule


class CelebADataModule(BaseDataModule):
    def setup(
            self,
            target_type: str = 'attr',
            val_ratio: float = 0.1,
            new_size: Tuple[int, int] = (256, 256),
            download: bool = False,
        ):
        data_transforms = Compose([
            Resize((256, 256)),
            ToTensor(),
            Normalize([0.5, 0.5, 0.5], [0.5, 0.5, 0.5]),
        ])
        full_dataset = CelebADataset(
            root=self.data_path,
            target_type=target_type,
            transform=data_transforms,
            download=download,
        )
        self.train_dataset = full_dataset


if __name__ == '__main__':
    dm = CelebADataModule(
        data_path='./data/celeba',
        batch_size=64,
        num_workers=4,
        download=True,
    )
    dm.setup()
    dl = dm.train_dataloader()
    print(len(dl))




import abc
from abc import ABC
from pathlib import Path

import torch
from torch import Tensor
from torch.utils.data import Dataset, DataLoader


class BaseDataModule(ABC):
    def __init__(
            self,
            data_path: Path,
            batch_size: int,
            num_workers: int,
        ):
        self.data_path = data_path
        self.batch_size = batch_size
        self.num_workers = num_workers

    @staticmethod
    def prepare_data(
            data_path: Path,
        ):
        raise NotImplementedError

    @abc.abstractmethod
    def setup(
            self,
            val_ratio: float,
        ) -> None:
        raise NotImplementedError

    def train_dataloader(self) -> DataLoader:
        train_dataloader = DataLoader(
            dataset=self.train_dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
        )

        return train_dataloader

    def val_dataloader(self) -> DataLoader:
        val_dataloader = DataLoader(
            dataset=self.val_dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
        )

        return val_dataloader

    def test_dataloader(self):
        pass




from typing import Optional

import numpy as np
import scipy.linalg as linalg
import torch
from torch import Tensor
from torch.nn import Module
from torch.utils.data import DataLoader
from torchvision.models import inception_v3 as Inception


class MetricCalculator:
    @staticmethod
    def calculate_frechet_distance(
            mu_1: Tensor,
            mu_2: Tensor,
            sigma_1: Tensor,
            sigma_2: Tensor,
            eps: float = 1e-6,
        ):
        mu_1 = mu_1.numpy()
        mu_2 = mu_2.numpy()
        sigma_1 = sigma_1.numpy()
        sigma_2 = sigma_2.numpy()

        delta_mu = mu_1 - mu_2

        offset = np.eye(sigma_1.shape[0]) * eps

        covmean, _ = linalg.sqrtm(
            (sigma_1 + offset) @ (sigma_2 + offset),
            disp=False,
        )

        if np.iscomplexobj(covmean):
            covmean = covmean.real

        distance = (
            delta_mu @ delta_mu +
            sigma1.trace() +
            sigma2.trace() -
            2 * covmean.trace()
        )

        return distance

    @staticmethod
    def calculate_activation_statistics(
            model,
            dataloader,
            inception,
        ):

        device = torch.device(
            'cuda' if next(a.parameters()).is_cuda else
            'cpu',
        )
        true_vecs, ae_vecs = [], []

        for images, _ in dataloader:
            bs = len(images)
            images = images.to(device)
            true_vecs.append(inception(images)[0].reshape(bs, -1))
            ae_vecs.append(inception(model.generate(images))[0].reshape(bs, -1))

        true_data = torch.stack(
            true_vecs,
        ).detach().cpu().view(bs * len(dataloader), -1).numpy()
        ae_data = torch.stack(
            ae_vecs,
        ).detach().cpu().view(bs * len(dataloader), -1).numpy()
        mu_1 = true_data.mean(axis=1)
        mu_2 = ae_data.mean(axis=1)

        sigma_1 = np.cov(true_data, rowvar=False)
        sigma_2 = np.cov(ae_data, rowvar=False)

        return mu_1, mu_2, sigma_1, sigma_2

    @classmethod
    @torch.no_grad()
    def calculate_fid(
            cls,
            model: Module,
            dataloader: DataLoader,
            classifier: Optional[Module] = None,
        ):
        if classifier is None:
            classifier = Inception(
                pretrained=True,
                progress=True,
            )
        model.eval()
        classifier.eval()

        mu_1, mu_2, sigma_1, sigma_2 = cls.calculate_activation_statistics(
            model=model,
            dataloader=dataloader,
            inception=classifier,
        )
        fid_value = cls.calculate_frechet_distance(
            mu_1=mu_1,
            mu_2=mu_2,
            sigma_1=sigma_1,
            sigma_2=sigma_2,
        )

        return fid_value

    @classmethod
    def criterion_adv(cls, val, targ):
        return 0.5 * torch.mean((val - targ) ** 2)

    @staticmethod
    def cov(m):
        fact = 1.0 / (m.size(1) - 1)
        m -= torch.mean(m, dim=1, keepdim=True)
        mt = m.t()
        return fact * m.matmul(mt).squeeze()

    @staticmethod
    def one_norm(p1, p2):
        return (p1 - p2).abs().sum()

    @classmethod
    def criterion_dist(cls, style):
        styles = style#torch.cat([orig_style, orig2_style])
        smeans = styles.mean()
        cov_m = cls.cov(styles)
        cov_diag = torch.diag(cov_m)

        device = style.device

        loss = (
            cls.one_norm(smeans, torch.ones(1).to(device)) +
            cls.one_norm(cov_m, torch.eye(cov_m.shape[0]).to(device)) +
            cls.one_norm(cov_diag, torch.ones(cov_diag.shape).to(device))
        )

        return loss




import random

import numpy as np
import torch


class Randomizer:
    @staticmethod
    def set_seed(seed: int = 9) -> None:
        torch.backends.cudnn.deterministic = True
        torch.manual_seed(seed)
        torch.cuda.manual_seed(seed)
        random.seed(seed)
        np.random.seed(seed)


if __name__ == '__main__':
    Randomizer.set_seed()




from torch.nn import (
    Module,
    Sequential,
)

from mr.blocks.conv_block import ConvBlock


class ResBlock(Module):
    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            kernel_size: int = 3,
            padding: int = 1,
        ):
        super().__init__()
        self.res_block = Sequential(
            ConvBlock(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=kernel_size,
                padding=padding,
                bias=False,
                instance_norm=True,
            ),
            ConvBlock(
                in_channels=out_channels,
                out_channels=out_channels,
                kernel_size=kernel_size,
                padding=padding,
                bias=False,
                instance_norm=True,
                act=False,
            ),
        )

    def forward(self, x):
        x = x + self.res_block(x)

        return x



from pathlib import Path
from pkg_resources import parse_requirements
from setuptools import find_packages, setup


def main():
    package_name = 'mr'
    root_dir = Path(__file__).parent

    with open(root_dir / 'README.md') as f:
        long_description = f.read()

    with open(root_dir / 'requirements.txt') as f:
        reqs = [str(req) for req in parse_requirements(f)]

    packages = find_packages(package_name)
    packages = list(map(lambda x: f'{package_name}/{x}', packages))

    setup(
        name=package_name,
        version='0.0.1',
        author='sergevkim',
        description=package_name,
        long_description=long_description,
        long_description_content_type='text/markdown',
        package_dir={package_name: package_name},
        packages=packages,
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
        ],
        python_requires='>=3.7',
        install_requires=reqs,
    )


if __name__ == '__main__':
    main()




from collections import OrderedDict

from torch.nn import (
    BatchNorm2d,
    Conv2d,
    ConvTranspose2d,
    InstanceNorm2d,
    MaxPool2d,
    Module,
    ReLU,
    Sequential,
)


class ConvBlock(Module):
    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            kernel_size: int = 3,
            stride: int = 1,
            padding: int = 1,
            bias: bool = True,
            transposed: bool = False,
            norm: bool = False,
            instance_norm: bool = False,
            pool: bool = False,
            act: bool = True,
        ):
        super().__init__()
        block_ordered_dict = OrderedDict()
        block_ordered_dict['conv'] = Conv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            bias=bias,
        ) if not transposed else ConvTranspose2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            bias=bias,
        )
        if norm:
            block_ordered_dict['norm'] = BatchNorm2d(num_features=out_channels)
        if instance_norm:
            block_ordered_dict['norm'] = InstanceNorm2d(
                num_features=out_channels,
                affine=True,
                track_running_stats=True,
            )
        if pool:
            block_ordered_dict['pool'] = MaxPool2d(kernel_size=2)
        if act:
            block_ordered_dict['act'] = ReLU()
        self.conv_block = Sequential(block_ordered_dict)

    def forward(self, x):
        x = self.conv_block(x)

        return x




from torch.nn import Module


class ParametersCounter:
    @staticmethod
    def count(
            model: Module,
            trainable: bool = False,
        ) -> int:
        if trainable:
            n_params = sum(
                p.numel() for p in model.parameters()
                if p.requires_grad
            )
        else:
            n_params = sum(p.numel() for p in model.parameters())

        return n_params

